#ifndef POPMIDI_H
#define POPMIDI_H

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include <math.h>

#include <x86intrin.h>

#include <SDL2/SDL.h>

#ifndef NDEBUG
#define ASSERT(expr) if (!(expr)) {*(int*)0 = 0;}
#else
#define ASSERT(expr)
#endif

#if 1
#define debug_printf(...) printf(__VA_ARGS__)
#else
#define debug_printf()
#endif

// Convenience macros
#define COUNT(array) (sizeof(array) / sizeof((array)[0]))
#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#define SQUARE(x) ((x)*(x))

#define KILOBYTES(n) ((size_t)1024LL*(n))
#define MEGABYTES(n) (1024LL*KILOBYTES(n))
#define GIGABYTES(n) (1024LL*MEGABYTES(n))
#define TERABYTES(n) (1024LL*GIGABYTES(n))

#define SIZEOF_POINTER (sizeof(void*))

#define pi32 3.14159265358979323846f
#define pi64 3.14159265358979323846
#define two_pi32 (2.0f*pi32)
#define two_pi64 (2.0*pi64)
#define inv_sqrt_2 0.70710678118f
#define deg_to_rad(x) ((x) * (pi32 / 180.f))

// Typedef choices for numerical types
typedef int8_t i8;
typedef int16_t i16;
typedef int32_t i32;
typedef int64_t i64;

typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;
typedef uint64_t u64;

typedef int32_t bool32;
typedef int8_t bool8;
#define true 1
#define false 0

typedef float f32;
typedef double f64;


typedef struct {
	u32 size;
	u8 data[0];
} file_buffer_type;

typedef struct midi_raw_chunk_type {
	char chunk_type[4];
	u32 chunk_length;
	union {
		struct {
			u16 format;
			u16 num_tracks;
			u16 time_division;
			u8 tracks[0];
		} header;
		u8 data[0];
	};

} midi_raw_chunk_type;

typedef struct midi_event_type {
	u32 delta_time;
	u8 event_type;
	union {
		struct {
			u8 channel;
			u8 param1;
			u8 param2;
		} channel;
		struct {
			u32 length;
			u8* data;
		} sysex;
		struct {
			u8 type;
			u32 length;
			u8* data;
		} meta;
	};

} midi_event_type;

typedef struct midi_track_type {
	u32 size;
	i32 num_events;
	midi_event_type* events;
	i32 event_index;
	i64 next_pause_tick;
} midi_track_type;

typedef struct parsed_midi_type {
	int num_tracks;
	midi_track_type* tracks;
	u32 ticks_per_beat;
} parsed_midi_type;

#pragma pack(push, 1)
typedef struct operator_type {
	u8 mul;
	u8 ksl_tl;
	u8 a_d;
	u8 s_r;
	u8 waveform;
} operator_type;

typedef struct instrument_type {
	u8 blocknum_low;
	u8 blocknum_high;
	u8 FB_conn;
	operator_type operators[2];
	u8 percussion;
	u8 unknown[2];
} instrument_type;
#pragma pack(pop)

// Useful wrapper to put a breakpoint on when debugging
static inline void
panic_and_exit() {
	exit(-1);
}

static inline bool32
is_power_of_2(u32 x) {
	bool32 result = x && ((x & (x - 1)) == 0);
	return result;
}

#endif // header guard
